/******************************************************************************
 * (c) Copyright 2002,2003, 1060 Research Ltd
 *
 * This Software is licensed to You, the licensee, for use under the terms of
 * the 1060 Public License v1.0. Please read and agree to the 1060 Public
 * License v1.0 [www.1060research.com/license] before using or redistributing
 * this software.
 *
 * In summary the 1060 Public license has the following conditions.
 * A. You may use the Software free of charge provided you agree to the terms
 * laid out in the 1060 Public License v1.0
 * B. You are only permitted to use the Software with components or applications
 * that provide you with OSI Certified Open Source Code [www.opensource.org], or
 * for which licensing has been approved by 1060 Research Limited.
 * You may write your own software for execution by this Software provided any
 * distribution of your software with this Software complies with terms set out
 * in section 2 of the 1060 Public License v1.0
 * C. You may redistribute the Software provided you comply with the terms of
 * the 1060 Public License v1.0 and that no warranty is implied or given.
 * D. If you find you are unable to comply with this license you may seek to
 * obtain an alternative license from 1060 Research Limited by contacting
 * license@1060research.com or by visiting www.1060research.com
 *
 * NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC
 * LICENSE V1.0 FOR DETAILS
 *
 * THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ
 * THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license
 *
 * File:          $RCSfile$
 * Version:       $Name$ $Revision$
 * Last Modified: $Date$
 *****************************************************************************/
package org.netkernel.layer0.util;

import org.w3c.dom.*;

import java.util.*;
/**
 * A Fast implementation XPath for simple expressions.
 * Evaluation shows it is an order of magnitude faster than xalan and will
 * work for about 2/3 of expressions in normal usage.
 * <br/>It will evaluate xpaths of the forms:
 * <ul><li> /a/b (absolute)
 * </li><li> a/b (relative)
 * </li><li> /a/* /c (disregard name)
 * </li><li> /a/b[1] (element position)
 * </li><li> /a/@b (attribute)
 * </li><li> //a (any element a below cursor)
 * </li></ul>
 * @author  tab
 */
public class FastXPath
{	/** all expressions are cached for quick evaluation */
	private static Map mCachedEvals = new HashMap(128);

	/** Inner class to store pre-parsed xpath expression */
	private static class SimpleEvalStruct
	{	String[] mNames;
		int[] mIndices;
		
		private static final int INDEX_ANY=-1;
		private static final int INDEX_ANYDEPTH=-2;
		
		/** construct/parse a new xpath expression */
		public SimpleEvalStruct(String aXPath)
		{	
			StringTokenizer stz = new StringTokenizer(aXPath,"/",true);
			int length = stz.countTokens();
			String[]  names = new String[length];
			int[] indices = new int[length];
			int c=0;
			int lastSlash=-2;
			boolean anydepth=false;
			for (int i=0; i<length; i++)
			{	String token = stz.nextToken();
				
				if (!token.equals("/"))
				{	
					//System.out.println("token of ["+token+"]");
					int i1 = token.indexOf('[');
					if (i1>=0)
					{	int i2 = token.indexOf(']');
						String indexString = token.substring(i1+1,i2);
						indices[c] = Integer.parseInt(indexString);
						names[c] = token.substring(0,i1);
					}
					else if (anydepth)
					{	indices[c]=INDEX_ANYDEPTH;
						names[c] = token;
					}
					else
					{	indices[c]=INDEX_ANY;
						names[c] = token;
					}
					if (names[c].equals("*"))
					{	names[c]=null;
					}
					c++;
					anydepth=false;
				}
				else
				{	if (lastSlash==i-1)
					{	//anydepth slash slash
						anydepth=true;
					}
					lastSlash=i;
					
				}
				
			}
			mNames = new String[c];
			System.arraycopy(names,0,mNames,0,c);
			mIndices = new int[c];
			System.arraycopy(indices,0,mIndices,0,c);
			
		}
		
		public List eval(Node aContextNode)
		{	List ns = new ArrayList();
			if (size()>0)
			{	evalDescend(ns,aContextNode,0);
			}
			else
			{	ns.add(aContextNode);
			}
			return ns;			
		}
		
		/** internal method to perform the iteration down the document
		 * @param aResult the nodeset to add final results too
		 * @param aCurrent the current node
		 * @param aLevels the parsed xpath structure
		 * @param aDepth the current depth
		 */
		private void evalDescend(List aResult, Node aCurrent, int aDepth)
		{	
			List ns = findMatch(aCurrent,aDepth);
			if (ns!=null)
			{	if (aDepth!=size()-1)
				{	for (int i=ns.size()-1; i>=0; i--)
					{	Node n = (Node)ns.get(i);
						evalDescend(aResult,n,aDepth+1);
					}
				}
				else
				{	aResult.addAll(ns);
				}
			}
		}			
		
		
		/** finds all nodes that match as the children of the given node at the
		 * given depth - will return null if none match
		 */
		public final List findMatch(Node aNode, int aDepth)
		{	List result=null;
			int foundIndex=0;
			String name = mNames[aDepth];
			int index = mIndices[aDepth];
			//SimpleEvalStruct qualifier = mQualifiers[aDepth];

			if (index==INDEX_ANYDEPTH)
			{	//anydepth search
				
				if (name!=null && name.length()>0 && name.charAt(0)=='@')
				{	// for attribute
					name = name.substring(1);
					Node n=XMLUtils.getFirstChildElement(aNode);
					while (n!=null)
					{	Element e=(Element)n;
						Node attr=e.getAttributeNode(name);
						if (attr!=null)
						{	if (result==null)
							{	result = new ArrayList(16);
							}
							result.add(n);
						}						
						
						n=XMLUtils.inOrderTraversalNext(n,aNode);
					}					
				}
				else
				{	// for elements
					Node n=XMLUtils.getFirstChildElement(aNode);
					while (n!=null)
					{	if (name==null || n.getNodeName().equals(name))
						{	if (result==null)
							{	result = new ArrayList(16);
							}
							result.add(n);
						}
						n=XMLUtils.inOrderTraversalNext(n,aNode);
					}
				}
			}
			else
			{	// child search
				if (name!=null && name.length()>0 && name.charAt(0)=='@')
				{	// for child attributes
					Node n=((Element)aNode).getAttributeNode(name.substring(1));
					if (n!=null)
					{	result = Collections.singletonList(n);
					}
				}
				else
				{	// for child elements
					for (Node n = aNode.getFirstChild(); n!=null; n=n.getNextSibling())
					{	if (n instanceof Element && ( name==null || n.getNodeName().equals(name)))
						{	foundIndex++;
							if (index==INDEX_ANY || foundIndex==index)
							{	if (result==null)
								{	result = new ArrayList(16);
								}
								result.add(n);
							}
						}
					}
				}
			}
			return result;
		}
		/** return the depth of the xpath expression */
		public final int size()
		{	return mNames.length;
		}
	}
	
	/** Returns a nodeset of matching nodes for the evaluation of the xpath
	 * at the given context node
	 */
	public static List eval(Node aContextNode, String aXPath) //throws TransformerException
	{	
		if (aXPath.length()>0 && aXPath.charAt(0)=='/' && !(aContextNode instanceof Document))
		{	aContextNode = aContextNode.getOwnerDocument();
		}		
		// parse simple xpath and cache
		SimpleEvalStruct levels = (SimpleEvalStruct)mCachedEvals.get(aXPath);
		if (levels==null)
		{	levels = new SimpleEvalStruct(aXPath);
			mCachedEvals.put(aXPath,levels);
		}
		return levels.eval(aContextNode);
	}
	
	public static Node getSingleNode(Node aContextNode, String aXPath) //throws TransformerException
	{	List ns = eval(aContextNode, aXPath);
		if (ns.size()==0)
		{	return null;
		}
		return (Node)ns.get(0);
	}
	
	public static boolean isTrue(Node aContextNode, String aXPath) //throws TransformerException
	{	List ns = eval(aContextNode, aXPath);
		return ns.size()>0;
	}
		
	/** Returns true if it is possible to evaluate the xpath using this engine
	 */
	public static boolean isSuitable(String aXPath)
	{	int length = aXPath.length();
		boolean simple=(length>0);
		char c1 = 0;
		for (int i=length-1; i>=0; i--)
		{	char c = aXPath.charAt(i);
			boolean validChar = (c>='a' && c<='z') || (c>='A' && c<='[') || (c>='/' && c<='9') || c==']' || c=='*' || c=='@';
			boolean badIndex = (c1==']' && (c<'0' || c<'9') );
			if (!validChar || badIndex )
			{	simple=false;
				break;
			}
			c1=c;
		}
		return simple;
	}
}